import os


try:
    from telegram.ext import Updater
    from telegram import ParseMode
except ImportError:
    Updater = None
    ParseMode = None

from autodora.observe import ProgressObserver


class TelegramObserver(ProgressObserver):
    def __init__(self):
        super().__init__()
        if Updater is None:
            raise RuntimeError("The TelegramObserver requires additional packages, please install them"
                               "(e.g. pip install autodora[telegram]).")
        try:
            self.updater = Updater(os.environ["TELEGRAM_BOT_TOKEN"])
            self.chat_id = os.environ["TELEGRAM_CHAT_ID"]
        except KeyError:
            raise RuntimeError("Please provide environment variables TELEGRAM_BOT_TOKEN and TELEGRAM_CHAT_ID")
        self.message_id = None
        self.total = None
        self.timed_out = 0
        self.done = 0
        self.errors = 0
        self.error_messages = []
        self.run_id = None
        self.run_platform = None

    def send_message(self, done=False):
        template = "*{run} ({start})*\n_{platform}_\n{done} succeeded," \
                   " {timed_out} timed out, {failed} failed{messages}"
        total_done = self.done + self.timed_out + self.errors
        message = template.format(
            start="{total_done} of {total}".format(total_done=total_done, total=self.total) if not done else "done",
            run=self.run_id,
            done=self.done,
            timed_out=self.timed_out,
            failed=self.errors,
            messages="",  # ("".join(["\n{}".format(m) for m in ["\nError messages:"] + self.error_messages])
            # if len(self.error_messages) > 0 else ""),
            platform=self.run_platform,
        )
        if self.message_id is None:
            self.message_id = self.updater.bot.send_message(chat_id=self.chat_id, text=message,
                                                            parse_mode=ParseMode.MARKDOWN).message_id
        else:
            print(message, type(message))
            self.updater.bot.edit_message_text(chat_id=self.chat_id, text=message, message_id=self.message_id,
                                               parse_mode=ParseMode.MARKDOWN)

    def run_started(self, platform, name, run_count, run_date, experiment_count):
        self.total = experiment_count
        self.run_id = "{}.{}".format(name, run_count)
        self.run_platform = platform
        self.send_message()

    def experiment_started(self, index, experiment):
        pass

    def experiment_finished(self, index, experiment):
        self.done += 1
        self.send_message()

    def experiment_interrupted(self, index, experiment):
        self.timed_out += 1
        self.send_message()

    def experiment_failed(self, index, experiment):
        self.errors += 1
        message = "[{}] {}".format(index, experiment["@error"].split("\n")[-2])
        self.error_messages.append(message)
        self.send_message()
        self.updater.bot.send_message(chat_id=self.chat_id, text=message)

    def run_finished(self, platform, name, run_count, run_date):
        self.send_message(True)



from argparse import ArgumentParser
from typing import Type, TYPE_CHECKING

from .analyze import is_excluded_from_string
from .runner import import_runner
from .storage import import_storage
from .analyze import add_arguments, show_from_args

if TYPE_CHECKING:
    from .experiment import Experiment


def parse_cli(cls):
    # type: (Type[Experiment]) -> None
    parser = ArgumentParser()
    parser.add_argument("storage")
    sub_parser = parser.add_subparsers(dest="mode")
    run_parser = sub_parser.add_parser("run")
    run_parser.add_argument("exp_id", type=int)

    analyze_parser = sub_parser.add_parser("analyze")
    analyze_parser.add_argument("-n", "--names", nargs="+", type=str)
    add_arguments(analyze_parser)

    explore_parser = sub_parser.add_parser("explore")
    explore_parser.add_argument("name", type=str, help="The name for this experiment")
    experiment = cls("")
    for parameter in experiment.parameters.parameters.values():
        explore_parser.add_argument("--{}".format(parameter.arg_name), type=parameter.p_type, nargs="+",
                                    help=parameter.description, dest="parameter.{}".format(parameter.arg_name))
    explore_parser.add_argument("-e", type=str, default=None, help="Execute the trajectory with the specified engine")
    explore_parser.add_argument("-t", type=int, default=None, help="Timeout for the execution")

    list_parser = sub_parser.add_parser("list")
    list_parser.add_argument("name", nargs="?", default=None)
    list_parser.add_argument("-e", "--exclude", type=str, default=None)

    remove_parser = sub_parser.add_parser("remove")
    remove_parser.add_argument("name")
    remove_parser.add_argument("-e", "--exclude", type=str, default=None)
    remove_parser.add_argument("--dry_run", action="store_true")

    groups_parser = sub_parser.add_parser("groups")

    # python product_experiment.py sqlite analyze

    args = parser.parse_args()
    storage = import_storage(args.storage)

    if args.mode == "run":
        exp_id = args.exp_id
        experiment = storage.get_experiment(cls, exp_id)
        experiment.run(True)
    elif args.mode == "analyze":
        experiments = []
        for name in args.names:
            for experiment in storage.get_experiments(cls, name):
                experiments.append(experiment)

        try:
            show_from_args(experiments, args)
        except KeyboardInterrupt:
            pass
    elif args.mode == "explore":
        from .trajectory import product, Trajectory
        settings = None
        for parameter in experiment.parameters.parameters.values():
            values = getattr(args, "parameter.{}".format(parameter.arg_name))
            if values:
                local_settings = {parameter.name: values}
                if settings is None:
                    settings = local_settings
                else:
                    settings = product(settings, local_settings)
        trajectory = Trajectory(args.name)
        trajectory.explore(cls, settings)
        print(*trajectory.experiments, sep="\n")
        if args.e:
            engine = import_runner(args.e, trajectory, storage, args.t)
            engine.run()
    elif args.mode == "list":
        if args.name:
            exclusion_filter = args.exclude
            print(*[e for e in storage.get_experiments(cls, args.name)
                    if not exclusion_filter or not is_excluded_from_string(exclusion_filter, e)], sep="\n")
        else:
            print(*storage.get_groups(), sep="\n")
    elif args.mode == "remove":
        exclusion_filter = args.exclude
        if exclusion_filter:
            for e in storage.get_experiments(cls, args.name):
                if not is_excluded_from_string(exclusion_filter, e):
                    storage.remove(args.name, experiment_id=e.identifier, dry_run=args.dry_run)
        else:
            storage.remove(args.name, dry_run=args.dry_run)



from datetime import datetime
import platform as platform_library
from typing import Optional

from .observe import ProgressObserver
from .storage import Storage
from .trajectory import Trajectory
from .runner import StoredRunner


class SimpleRunner(StoredRunner):
    def __init__(self, trajectory: Trajectory, storage: Storage, observer: Optional[ProgressObserver] = None,
                 repeat=False):
        super().__init__(trajectory, storage, observer, repeat)

    def run(self):
        run_date = datetime.now()
        platform = platform_library.node()
        name = self.trajectory.name
        experiments = [e for s, e in zip(self.trajectory.settings, self.trajectory.experiments)
                       if self.should_run(s, e)]
        if self.observer:
            experiment_count = len(self.trajectory.experiments)
            self.observer.run_started(platform, name, self.run_count, run_date, experiment_count)
        for i, experiment in enumerate(experiments):
            experiment.config["@run.count"] = self.run_count if self.storage else -1
            experiment.config["@run.date"] = run_date
            experiment.config["@run.computer"] = platform
            if self.storage:
                experiment.save(self.storage)

            if self.observer:
                self.observer.experiment_started(i, experiment)

            # noinspection PyBroadException
            try:
                experiment.run(bool(self.storage))
                if self.observer:
                    self.observer.experiment_finished(i, experiment)
            except Exception:
                if self.observer:
                    self.observer.experiment_failed(i, experiment)

        if self.observer:
            self.observer.run_finished(platform, name, self.run_count, run_date)
        return self.trajectory.experiments



import atexit
import errno
import multiprocessing
import os
import signal
import subprocess
from multiprocessing import Queue, Manager, Process
from multiprocessing.pool import Pool
from subprocess import TimeoutExpired
from traceback import print_exc
from typing import Optional, Union, Any

from temporary import temp_file

from .observe import Observer, dispatch


class Update:
    SENTINEL = "sentinel"
    STARTED = "started"
    DONE = "done"
    TIMEOUT = "timeout"
    FAILED = "failed"

    def __init__(self, status, index, command, meta):
        self.status = status
        self.index = index
        self.command = command
        self.meta = meta


class ParallelObserver(Observer):
    @dispatch
    def observe(self, update):
        # type: (Update) -> None
        raise NotImplementedError()


def monitor(filename, monitor_queue):
    with open(filename, "w") as ref:
        while True:
            update = monitor_queue.get()  # type: Union[Update, str]
            if update == Update.SENTINEL:
                return
            elif update.status == Update.STARTED:
                print("ADD", update.meta, sep=" ", file=ref)
            else:
                print("REM", update.meta, sep=" ", file=ref)


def observe(observer, queue, count=None):
    # type: (ParallelObserver, Queue, Optional[int]) -> None
    to_see = None if count is None else set(range(count))
    while to_see is None or len(to_see) > 0:
        update = queue.get()  # type: Union[Update, str]
        if update == Update.SENTINEL:
            return
        elif isinstance(update, Update):
            if update.status == Update.DONE or update.status == Update.TIMEOUT or update.status == Update.FAILED:
                if to_see:
                    to_see.remove(update.index)
            try:
                observer.observe(update)
            except Exception:
                print_exc()
        else:
            raise ValueError("Invalid update {}".format(update))


def run_command(command, timeout=None):
    return worker((-1, None, command, timeout, None, None))


def run_function(f, *args, timeout=None, **kwargs):
    worker((-1, None, (f, args, kwargs), timeout, None, None))


def worker(args):
    i, meta, command, timeout, queue, m_queue = args  # type: (int, Any, Any, int, Queue, Queue)
    # TODO Capture output?

    if isinstance(command, str):
        is_string = True
    else:
        try:
            is_string = len(command) > 0 and not callable(command[0])
        except TypeError:
            raise ValueError("Command must be either string, args or function-args pair")

    if is_string:
        with subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                              start_new_session=True) as process:
            try:
                if m_queue:
                    m_queue.put(Update(Update.STARTED, i, command, process.pid))
                if queue:
                    queue.put(Update(Update.STARTED, i, command, meta))
                out, err = process.communicate(timeout=timeout)
                if queue:
                    if process.returncode == 0:
                        queue.put(Update(Update.DONE, i, command, meta))
                    else:
                        queue.put(Update(Update.FAILED, i, command, meta))
                return out.decode(), err.decode()
            except TimeoutExpired:
                try:
                    os.killpg(process.pid, signal.SIGINT)  # send signal to the process group
                except OSError as e:
                    if e.errno != errno.ESRCH:
                        if e.errno == errno.EPERM:
                            os.waitpid(-process.pid, 0)
                    else:
                        raise e
                finally:
                    if queue:
                        queue.put(Update(Update.TIMEOUT, i, command, meta))

                process.communicate()
            finally:
                if m_queue:
                    m_queue.put(Update(Update.DONE, i, command, process.pid))

    else:
        assert isinstance(command, (tuple, list))
        if len(command) < 2:
            command = command + ([],)
        if len(command) < 3:
            command = command + (dict(),)

        f, args, kwargs = command

        p = multiprocessing.Process(target=f, args=args, kwargs=kwargs)
        if queue:
            queue.put(Update(Update.STARTED, i, command, meta))

        p.start()
        p.join(timeout)

        if p.is_alive():
            p.terminate()
            p.join()
            if queue:
                queue.put(Update(Update.TIMEOUT, i, command, meta))

        else:
            if queue:
                queue.put(Update(Update.DONE, i, command, meta))


def status(s):
    """Prints things in bold."""
    print('\033[1m{0}\033[0m'.format(s))


def run_commands(commands, processes=None, timeout=None, meta=None, observer=None):
    pool = Pool(processes=processes)
    manager, queue, m = None, None, None
    manager = Manager()
    m = manager.Queue()
    if observer:
        queue = manager.Queue()

    if meta:
        commands = [(i, meta, command, timeout, queue, m) for i, (command, meta) in enumerate(zip(commands, meta))]
    else:
        commands = [(i, meta, command, timeout, queue, m) for i, command in enumerate(commands)]

    with temp_file() as f:
        filename = f

    m_process = Process(target=monitor, args=(filename, m))
    m_process.daemon = True
    m_process.start()

    def clean_exit():
        status("Keyboard interrupt intercepted, shutting down")
        try:
            m_process.terminate()
            m_process.join()
        except Exception:
            status("Monitor process could not be shut down")
            print_exc()

        try:
            pool.terminate()
            pool.join()
        except Exception:
            status("Pool could not be shut down")
            print_exc()

        status("Shutting down potential orphan processes")
        active = set()
        with open(filename) as ref:
            for line in ref:
                parts = line.split(" ")
                if parts[0] == "ADD":
                    active.add(int(parts[1]))
                elif parts[0] == "REM":
                    active.remove(int(parts[1]))

        for pid in active:
            try:
                print("Killing", pid)
                os.killpg(pid, signal.SIGINT)  # send signal to the process group
            except OSError as e:
                if e.errno != errno.ESRCH:
                    if e.errno == errno.EPERM:
                        os.waitpid(-pid, 0)
                else:
                    raise e
            except Exception:
                print_exc()
                pass

        os.unlink(filename)

        status("Completely shut down")

    r = pool.map_async(worker, commands)
    atexit.register(clean_exit)

    if observer:
        observe(observer, queue, len(commands))

    r.wait()
    status("### DONE ##")
    m.put(Update.SENTINEL)
    m_process.join()




from typing import Dict, Any, List


def flatten(dict_settings):
    common_length = None
    for val in dict_settings.values():
        if not common_length:
            common_length = len(val)
        elif len(val) != common_length:
            raise ValueError("All value assignments must have the same length")
    return [{k: v[i] for k, v in dict_settings.items()} for i in range(common_length)]


def product(*args):
    args = [flatten(arg) if isinstance(arg, dict) else arg for arg in args]
    result = args[0]
    for arg in args[1:]:
        result = [dict(**d1, **d2) for d1 in result for d2 in arg]
    return result


class Trajectory(object):
    def __init__(self, name):
        self.experiments = []
        self.settings = []
        self.name = name

    def add(self, experiment):
        self.experiments.append(experiment)

    def explore(self, cls, settings: List[Dict[str, Any]]):
        if isinstance(settings, dict):
            common_length = None
            for val in settings.values():
                if not common_length:
                    common_length = len(val)
                elif len(val) != common_length:
                    raise ValueError("All value assignments must have the same length")
            if common_length is None:
                settings = [{}]
            else:
                settings = [{k: v[i] for k, v in settings.items()} for i in range(common_length)]

        for setting in settings:
            self.settings.append(setting)
            experiment = cls(self.name)
            for key, value in setting.items():
                experiment[key] = value

            self.add(experiment)



import inspect
import platform as platform_library
from datetime import datetime
from typing import TYPE_CHECKING

from .observe import ProgressObserver
from .parallel import ParallelObserver, Update
from . import parallel
from .storage import export_storage

if TYPE_CHECKING:
    from storage import Storage


class ParallelToProcess(ParallelObserver):
    def __init__(self, observer, runner):
        # type: (ProgressObserver, CommandLineRunner) -> None
        super().__init__()
        self.observer = observer
        self.runner = runner

    def observe(self, update):
        if self.observer.auto_load:
            meta = self.runner.trajectory.experiments[update.index]
            if update.status == Update.DONE or update.status == Update.FAILED:
                meta = meta.fresh_copy()
        else:
            meta = update.meta

        if update.status == Update.STARTED:
            self.observer.experiment_started(update.index, meta)
        if update.status == Update.DONE:
            self.observer.experiment_finished(update.index, meta)
        if update.status == Update.TIMEOUT:
            self.observer.experiment_interrupted(update.index, meta)
        if update.status == Update.FAILED:
            self.observer.experiment_failed(update.index, meta)


class PrintObserver(ProgressObserver):
    def __init__(self, auto_load=True):
        super().__init__(auto_load=auto_load)
        self.experiment_count = None

    def run_started(self, platform, name, run_count, run_date, experiment_count):
        self.experiment_count = experiment_count
        print("[{}] started: {} - {}".format(platform, name, run_count))

    def experiment_started(self, index, experiment):
        print("[{}] started: {}".format(index, experiment))

    def experiment_finished(self, index, experiment):
        print("[{}] done: {}".format(index, experiment))

    def experiment_interrupted(self, index, experiment):
        print("[{}] timed out: {}".format(index, experiment))

    def experiment_failed(self, index, experiment):
        print("[{}] failed: {}".format(index, experiment))

    def run_finished(self, platform, name, run_count, run_date):
        print("[{}] done: {} - {}".format(platform, name, run_count))


class Runner:
    def __init__(self, trajectory, observer):
        self.trajectory = trajectory
        self.observer = observer

    @staticmethod
    def setting_exists(setting, previous_experiments):
        for e in previous_experiments:
            for k, v in setting.items():
                if e[k] != v:
                    return True
        return False


class StoredRunner(Runner):
    def __init__(self, trajectory, storage, observer, repeat):
        super().__init__(trajectory, observer)
        self.storage = storage  # type: Storage
        self.repeat = repeat
        self._previous_experiments = None
        self.run_count = None if storage is None else self.storage.get_new_run()

    def setting_exists(self, setting, experiment):
        if self.storage is None:
            return False
        if self._previous_experiments is None:
            self._previous_experiments = self.storage.get_experiments(experiment.__class__, self.trajectory.name)
        return Runner.setting_exists(setting, self._previous_experiments)

    def should_run(self, setting, experiment):
        if not self.repeat and self.setting_exists(setting, experiment):
            return False
        return True


class CommandLineRunner(StoredRunner):
    def __init__(self, trajectory, storage, processes=None, timeout=None, observer=None, via_cli=True, repeat=False):
        super().__init__(trajectory, storage, None if observer is None else ParallelToProcess(observer, self), repeat)
        self.timeout = timeout
        self.processes = processes
        self.via_cli = via_cli

    def run(self):
        commands = []
        run_date = datetime.now()
        platform = platform_library.node()
        name = self.trajectory.name

        experiments = [e for s, e in zip(self.trajectory.settings, self.trajectory.experiments)
                       if self.should_run(s, e)]

        if self.observer:
            experiment_count = len(experiments)
            self.observer.observer.run_started(platform, name, self.run_count, run_date, experiment_count)

        for experiment in experiments:
            if self.timeout:
                experiment.config["@timeout"] = self.timeout
            experiment.config["@run.count"] = self.run_count
            experiment.config["@run.date"] = run_date
            experiment.config["@run.computer"] = platform
            experiment.save(self.storage)
            storage_name = export_storage(experiment.storage)
            cls = experiment.__class__
            filename = inspect.getfile(cls)
            if self.via_cli:
                commands.append("python {} {} run {}".format(filename, storage_name, experiment.identifier))
            else:
                commands.append((CommandLineRunner.run_single, (self.storage, cls, experiment.identifier)))

        meta = [e.identifier for e in experiments] if self.observer else None
        parallel.run_commands(commands, timeout=self.timeout, observer=self.observer, meta=meta,
                              processes=self.processes)
        if self.observer:
            self.observer.observer.run_finished(platform, name, self.run_count, run_date)
        return [self.storage.get_experiment(e.__class__, e.identifier) for e in self.trajectory.experiments]

    @staticmethod
    def run_single(storage, cls, identifier):
        experiment = storage.get_experiment(cls, identifier)
        experiment.run(True)
        return experiment.identifier


def import_runner(runner_string, trajectory, storage, timeout=None):
    if runner_string == "cli":
        return CommandLineRunner(trajectory, storage, timeout=timeout)
    elif runner_string == "multi":
        return CommandLineRunner(trajectory, storage, timeout=timeout, via_cli=False)
    else:
        raise ValueError("Could not parse runner from {}".format(runner_string))



import time
import traceback
from datetime import datetime
from typing import Union, Any, Dict, List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from storage import Storage

from .trajectory import Trajectory


class Parameter(object):
    def __init__(self, p_type, default=None, description=None, name=None, arg_name=None):
        self.name = name
        self.p_type = p_type
        self.default = default
        self.description = description
        self.specific_arg_name = arg_name

    @property
    def arg_name(self):
        return self.specific_arg_name or self.name


class Result(Parameter):
    pass


class Config(Parameter):
    pass


class Group(object):
    def __init__(self, name):
        self.name = name
        self.parameters = dict()  # type: Dict[str, Parameter]
        self.values = dict()  # type: Dict[str, Any]

    def add_parameter(self, name, p_type, default=None, description=None, arg_name=None):
        self.add(Parameter(p_type, default, description, name, arg_name))

    def add(self, parameter):
        self.parameters[parameter.name] = parameter

    def set_values(self, **kwargs):
        for key, value in kwargs.items():
            self.set_value(key, value)

    def set_value(self, name, value):
        if name not in self.parameters:
            raise ValueError("No parameter called {name}".format(name=name))
        if not isinstance(value, self.parameters[name].p_type):
            raise ValueError("Unexpected type for value {value}".format(value=value))
        self.values[name] = value

    def __setitem__(self, key, value):
        self.set_value(key, value)

    def __getitem__(self, item):
        return self.values[item] if item in self.values else self.parameters[item].default

    def add_arguments(self, parser, prefix=None):
        for name, parameter in self.parameters.items():
            parser.add_argument(
                "--{}{}".format(prefix or '', parameter.arg_name),
                type=parameter.p_type,
                default=parameter.default,
                help=parameter.description
            )

    def parse_arguments(self, args, prefix=None):
        for name, parameter in self.parameters.items():
            self.set_value(name, getattr(args, "{}{}".format(prefix or '', parameter.arg_name)))

    def get_arguments(self, prefix=None):
        return " ".join(
            "--{} {}".format("{}{}".format(prefix or '', self.parameters[name].arg_name), value)
            for name, value in self.values.items() if value is not None
        )

    def copy(self):
        group = self.__new__(self.__class__)
        group.parameters = dict(self.parameters)
        group.values = dict(self.values)
        return group

    def __str__(self):
        values = ", ".join("{}: {}".format(key, self.values.get(key, '')) for key in self.parameters)
        return "{}={{{}}}".format(self.name, values)


class Derived(object):
    def __init__(self, callback, cache):
        self.callback = callback
        self.cache = cache

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)


def derived(cache):
    def real(func):
        return Derived(func, cache)
    return real


class Experiment(object):
    def __init__(self, group, storage=None, identifier=None):
        # type: (str, Optional[Storage], Optional[int]) -> None
        self.derived_callbacks = dict()
        self.storage = storage
        self.identifier = identifier

        self.group = group

        self.config = Group("config")
        self.config.add_parameter("@timeout", int, None, "The timeout value set for this experiment")
        self.config.add_parameter("@run.count", int, None, "The run count (for local storage)")
        self.config.add_parameter("@run.computer", str, None, "The computer name the run was performed on")
        self.config.add_parameter("@run.date", datetime, None, "The date when the run was instantiated")

        self.parameters = Group("parameters")

        self.result = Group("result")
        self.result.add_parameter("@error", str, None, "Potential error messages")
        self.result.add_parameter("@start_time", datetime, None, "When this experiment was started")
        self.result.add_parameter("@end_time", datetime, None, "When this experiment was started")
        self.result.add_parameter("@runtime", float, None, "How long the experiment took to execute (perf time)")
        self.result.add_parameter("@runtime_wall", float, None,
                                  "How long the experiment took to execute (wall clock time)")
        self.result.add_parameter("@runtime_process", float, None,
                                  "How long the experiment took to execute (process time)")

        self.derived_callbacks["@completed"] = Derived(self.is_completed, False)

        self.derived = dict()

        for key, value in self.__class__.__dict__.items():
            if isinstance(value, Parameter):
                if value.name is None:
                    value.name = key
                if isinstance(value, Config):
                    self.config.add(value)
                elif isinstance(value, Result):
                    self.result.add(value)
                else:
                    self.parameters.add(value)
            elif isinstance(value, Derived):
                if key.startswith("derived_"):
                    key = key[8:]
                self.derived_callbacks[key] = value

    def is_completed(self):
        return self["@end_time"] is not None

    def get_derived(self, name):
        if name in self.derived:
            return self.derived[name]

        if name in self.derived_callbacks:
            try:
                result = self.derived_callbacks[name](self)
            except TypeError:
                result = self.derived_callbacks[name]()

            if self.derived_callbacks[name].cache:
                self.derived[name] = result
            return result

        raise ValueError("There is no derived attribute with the name {name}".format(name=name))

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, value):
        self.set(key, value)

    def get(self, name: Union[str, Parameter]):
        if isinstance(name, Parameter):
            name = name.name
        parts = name.split(".", 1)
        if parts[0] == "par" or parts[0] == "parameter":
            return self.parameters[parts[1]]
        elif parts[0] == "res" or parts[0] == "result":
            return self.result[parts[1]]
        elif parts[0] == "conf" or parts[0] == "config":
            return self.config[parts[1]]
        elif parts[0] == "derived":
            return self.get_derived(parts[1])
        else:
            name = ".".join(parts)
            results = []
            if name in self.config.parameters:
                results.append(self.config[name])
            elif name in self.parameters.parameters:
                results.append(self.parameters[name])
            elif name in self.result.parameters:
                results.append(self.result[name])
            elif name in self.derived_callbacks:
                results.append(self.get_derived(name))
            if len(results) == 1:
                return results[0]
            elif len(results) > 1:
                raise ValueError("Multiple entries found for the name {name}, please use parameter.{name}, "
                                 "result.{name}, config.{name} or derived.{name} to disambiguate".format(name=name))
        raise ValueError("No entry found for the name {name}".format(name=name))

    def set(self, name, value):
        if isinstance(name, Parameter):
            name = name.name

        parts = name.split(".", 1)
        if parts[0] == "par" or parts[0] == "parameter":
            self.parameters[parts[1]] = value
        elif parts[0] == "res" or parts[0] == "result":
            self.result[parts[1]] = value
        elif parts[0] == "conf" or parts[0] == "config":
            self.config[parts[1]] = value
        elif parts[0] == "derived":
            self.get_derived(parts[1])
        else:
            name = ".".join(parts)
            results = []
            if name in self.config.parameters:
                results.append(self.config)
            elif name in self.parameters.parameters:
                results.append(self.parameters)
            elif name in self.result.parameters:
                results.append(self.result)

            if len(results) == 1:
                results[0][name] = value
            elif len(results) > 1:
                raise ValueError("Multiple entries found for the name {name}, please use parameter.{name}, "
                                 "result.{name}, config.{name} or derived.{name} to disambiguate".format(name=name))
            else:
                raise ValueError("No entry found for the name {name}".format(name=name))

    def run(self, auto_save=True):
        try:
            self.result["@start_time"] = datetime.now()
            if auto_save:
                self.save()
            start = time.perf_counter()
            start_process = time.process_time()
            start_wall = time.time()
            self.run_internal()
            runtime = time.perf_counter() - start
            runtime_process = time.process_time() - start_process
            runtime_wall = time.time() - start_wall
            self.result["@end_time"] = datetime.now()
            self.result["@runtime"] = runtime
            self.result["@runtime_process"] = runtime_process
            self.result["@runtime_wall"] = runtime_wall
            if auto_save and self.storage:
                self.save()
            return self
        except Exception:
            self.result["@error"] = "ERROR\n" + traceback.format_exc()
            if auto_save and self.storage:
                self.save()
            raise

    def run_internal(self):
        raise NotImplementedError()

    def save(self, storage=None):
        if storage:
            storage.save(self)
        else:
            if not self.storage:
                raise ValueError("No storage specified")
            self.storage.save(self)

    def fresh_copy(self):
        return self.storage.get_experiment(self.__class__, self.identifier)

    def __str__(self):
        return "EXP(group={}, id={}, {}, {}, {}, derived={})"\
            .format(self.group or '', self.identifier or '', self.config, self.parameters, self.result, self.derived)

    def add_arguments(self, parser, prefix=None):
        self.config.add_arguments(parser, "{}conf.".format(prefix or ''))
        self.parameters.add_arguments(parser, "{}par.".format(prefix or ''))
        self.result.add_arguments(parser, "{}res.".format(prefix or ''))

    def parse_arguments(self, args, prefix=None):
        self.config.parse_arguments(args, "{}conf.".format(prefix or ''))
        self.parameters.parse_arguments(args, "{}par.".format(prefix or ''))
        self.result.parse_arguments(args, "{}res.".format(prefix or ''))

    def get_arguments(self, prefix=None):
        strings = [
            self.config.get_arguments("{}conf.".format(prefix or '')),
            self.parameters.get_arguments("{}par.".format(prefix or '')),
            self.result.get_arguments("{}res.".format(prefix or '')),
        ]
        return " ".join(strings)

    @classmethod
    def explore(cls, name: str, settings: Union[List[Dict[str, Any]], Dict[str, List]]):
        trajectory = Trajectory(name)
        trajectory.explore(cls, settings)
        return trajectory

    @classmethod
    def enable_cli(cls):
        if cls.__module__ == "__main__":
            cls.run_cli()

    @classmethod
    def run_cli(cls):
        from .cli import parse_cli
        parse_cli(cls)

    def __getstate__(self):
        return {
            "group": self.group,
            "identifier": self.identifier,
            "storage": self.storage,
            "config": self.config,
            "parameters": self.parameters,
            "result": self.result,
            "derived": self.derived,
        }

    def __setstate__(self, state):
        self.__init__(state["group"], state["storage"], state["identifier"])
        self.config = state["config"]
        self.parameters = state["parameters"]
        self.result = state["result"]
        self.derived = state["derived"]



import os
import shutil
import sys

from setuptools import setup, find_packages, Command
from os import path

NAME = 'autodora'
DESCRIPTION = 'Automate experiments and explore your data.'
URL = 'http://github.com/samuelkolb/autodora'
EMAIL = 'samuel.kolb@me.com'
AUTHOR = 'Samuel Kolb'
REQUIRES_PYTHON = '>=3.5.0'
VERSION = "0.3.19"

# What packages are required for this module to be executed?
REQUIRED = [
    "matplotlib", "peewee", "pebble", "numpy", "temporary"
]

# What packages are optional?
EXTRAS = {
        "telegram": ["python-telegram-bot"]
}

# Distribute: python setup.py upload
# Requires: pip install twine wheel

here = os.path.abspath(os.path.dirname(__file__))

with open(path.join(here, "README.md")) as ref:
    long_description = ref.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            shutil.rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        # self.status('Pushing git tags…')
        # os.system('git tag v{0}'.format(about['__version__']))
        # os.system('git push --tags')

        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(exclude=('tests',)),
    zip_safe=False,
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    setup_requires=['pytest-runner'],
    tests_require=["pytest"],
    entry_points={},
    cmdclass={
        'upload': UploadCommand,
    },
)